/* RESETING DEFAULT MARGINS AND PADDINGS FOR ALL ELEMENTS ------------------------------- */
* {
    margin: 0;
    padding: 0;
    box-sizing: border-box; /* margins and paddings included in width and height of elements */
}

.stApp {
    /* background-color: #F4F4F4; */
    background-color: #ebebeb;
    font-family: 'Inter', 'Segoe UI', 'Roboto', sans-serif;
    color: darkgray !important;
}

/* SETTINGS FOR THE MAIN CONTAINER OF THE DASHBOARD ------------------------------- */
.block-container {
    min-width: 600px;
    width: 90% !important;
    margin-left: auto !important;
    margin-right: auto !important;
    padding-left: 2rem !important;
    padding-right: 2rem !important;
}

/*
.block-container {
    padding-left: 2rem !important;
    padding-right: 2rem !important;
    max-width: 100% !important;
    margin: 0 !important;
}
*/


.stSubheader {
    color: #6C929E !important;
}
.st-key-align{
    display: flex !important;
    align-items: stretch !important;
    gap: 20px;
}

.dashboard-title { 
    display: flex; 
    justify-content: center; 
    align-items: center; 
    width: 100%; 
    max-width: 100vw; 
    height: 65px; 
    background-color: transparent; /* przezroczyste tło */
    color: #426c85;
    font-family: 'Inter', sans-serif; 
    font-size: 38px; 
    font-weight: 500; 
    letter-spacing: 0.5px; 
    position: sticky; 
    top: 0; 
    z-index: 100; 
    box-sizing: border-box; 
    line-height: 1; 
    margin: 0 0 1rem 0 !important; 
    padding: 0 !important; 
    border: none;
    border-bottom: 1.5px solid #426c85;
    border-top: 1.5px solid #426c85;
    border-radius: 0;
}

/*  WHITE CONTAINERS for all elments ----------------------------------------------------------------- */

.st-key-table_container,.st-key-basics, .st-key-ratio_white_container,
.st-key-proba-probabilities,.st-key-capacity-per-sample, .st-key-pareto, .st-key-gauge {
    background-color: white;
    border-radius: 8px;
    box-shadow: 0 6px 10px rgba(0,0,0,0.1);
    padding: 8px 8px;
    border: 1px solid #e6e6e6;
}

/*  WHITE CONTAINERS for all elments : More padding ----------------------------------------------------------------- */

.st-key-ratio_epsilon_white_container,.st-key-pattern_white_container,.st-key-pattern_epsilon_white_container,
.st-key-ambig_epsilon_container,.st-key-discr_epsilon_container,.st-key-lolipop_container,
.st-key-capacity-hist,.st-key-capacity-boxplot, 
.st-key-percent_agreement,.st-key-agreement_rates_violin, .st-key-vprs_by_class, .st-key-vprs_hist, .st-key-vprs_vs_base_model,
.st-key-cohens_kappa_barplot{
    background-color: white;
    border-radius: 8px;
    box-shadow: 0 6px 10px rgba(0,0,0,0.1);
    padding: 8px 8px;
    border: 1px solid #e6e6e6;
    padding-bottom: 25px;
}

.st-key-venn{
    background-color: white;
    border-radius: 8px;
    box-shadow: 0 6px 10px rgba(0,0,0,0.1);
    border: 1px solid #e6e6e6;
    padding-bottom: 152px;
    padding-left:0px;
    padding-right: 0px;

}

/*  WHITE CONTAINERS for all elments : Other classes ----------------------------------------------------------------- */

.st-key-per_sample{
    background-color: white;
    border-radius: 8px;
    box-shadow: 0 6px 10px rgba(0,0,0,0.1);
    padding: 8px 8px;
    border: 1px solid #e6e6e6;
    padding-bottom:22px;
  

}

.st-key-fi_table{
    background-color: white;
    border-radius: 8px;
    box-shadow: 0 6px 10px rgba(0,0,0,0.1);
    padding: 8px 8px;
    border: 1px solid #e6e6e6;
    padding-bottom:28px;
    margin-top: 0px;
}
.st-key-fi_heatmap{
    background-color: white;
    border-radius: 8px;
    box-shadow: 0 6px 10px rgba(0,0,0,0.1);
    padding: 8px 8px;
    border: 1px solid #e6e6e6;
    padding-bottom:50px;
    height: 700px
}

.st-key-capacity_per_sample_padding{
    margin-top:15px;
}


/* -------SECTIONS TITLES STYLE----------------------------------------------------- */

.section_title{
    color: #426c85;
    font-family: 'Inter', sans-serif; 
    font-size: 21px; 
    font-weight: 500; 
    letter-spacing: 0.5px; 
    margin-bottom:10px;
   
}

/* -------SECTIONS DESCRIPTIONS STYLE----------------------------------------------------- */

.section_description{
    color: #6B7F8A;
    font-family: 'Inter', sans-serif; 
    font-size: 15px; 
    font-weight: 300 !important;
    letter-spacing: 0.1px;
    line-height: 1.6; 
    margin-bottom:15px;
   
}
.venn_description{
    color: #6B7F8A;
    font-family: 'Inter', sans-serif; 
    font-size: 15px; 
    font-weight: 300 !important;
    letter-spacing: 0.1px;
    line-height: 1.6; 
    margin-bottom:30px;
    padding: 8px 8px;
   
}

.less_padding{
    margin-bottom: 10px;
}

.no_padding{
    margin-bottom: 0px;
    padding: 0px 0px;
}

.additional_padding{
    margin-bottom: 35px;
}

.heatmap-descr{
    padding: 10px 10px;
    margin-top: 10px;
    margin-bottom: 15px;
}

/* ------- PLOTS DESCRIPTIONS STYLE----------------------------------------------------- */

.plots_descr{
    font-family: 'Inter', 'Segoe UI', sans-serif;
    font-weight: 300;
    line-height: 1.2;
    font-size: 15px;
    color: #989898;
}

.rashomon-table-descr {
    padding: 8px 10px;
    margin-bottom: 15px;
    margin-top: 0px;
}

/* ------- RADIO BUTTON STYLE ----------------------------------------------------- */


div[role="radiogroup"] {
    gap: 10px;
    margin: 0 !important;
    padding: 0 !important;
}

div[role="radiogroup"] > label {
    margin-bottom: 0 !important;
    margin-top: 0 !important;
    background: transparent !important;
}
div[role="radiogroup"] > label[data-testid="stRadio"] {
    margin: 0 !important;
}

div[role="radiogroup"] p { 
    color: #8B9DA8; 
    font-family: 'Inter', sans-serif; 
    font-size: 16px; 
    font-weight: 300 !important; 
    letter-spacing: 0.1px;
    line-height: 1.5; 
    margin-bottom: 0px !important;
}

div[role="radiogroup"] input[type="radio"] {
    display: none;
}